# -*- mode: python -*-
try:
    from importlib.metadata import version

    __version__ = version("dlab")
except Exception:
    # If package is not installed (e.g. during development)
    __version__ = "unknown"



# -*- mode: python -*-
"""Functions for using kilosort/phy data"""

import json
import logging
import re
from collections.abc import Iterable, Iterator
from pathlib import Path
from typing import NamedTuple

import arf
import ewave
import h5py as h5
import numpy as np
import pandas as pd
import quickspikes as qs
import toelis

from dlab import neurobank as nbank
from dlab import pprox
from dlab.spikes import SpikeWaveforms, save_waveforms

log = logging.getLogger(__name__)


class Trial(NamedTuple):
    """Represents the structure of a trial. All time units are in samples."""

    recording_entry: int
    recording_start: int
    recording_end: int
    stimulus_name: str
    stimulus_start: int
    stimulus_end: int


class Stimulus(NamedTuple):
    name: str
    start: int
    end: int | None = None


def read_kilo_params(fname: Path) -> dict:
    """Read the kilosort params.py file"""
    from configparser import ConfigParser
    from itertools import chain

    parser = ConfigParser()
    with open(fname) as lines:
        lines = chain(("[top]",), lines)
        parser.read_file(lines)
    sect = parser["top"]
    return dict(
        dtype=sect["dtype"].strip("'"),
        nchannels=int(sect["n_channels_dat"]),
        sampling_rate=float(sect["sample_rate"]),
    )


def oeaudio_stims(dset: h5.Dataset) -> Iterator[Stimulus]:
    """Parse the messages in the 'stim' dataset to get a table of stimuli with
    start samples. Note that these will need to be corrected for offset of the
    recording and network lag.

    """
    re_start = re.compile(r"start (.*)")
    for row in dset:
        time = row["start"]
        message = row["message"].decode("utf-8")
        m = re_start.match(message)
        if m is not None:
            stim_name = Path(m.group(1)).stem
            yield Stimulus(stim_name, time)


def entry_time(entry):
    """Return the timestamp of an entry as a floating point number"""
    from arf import timestamp_to_float

    return timestamp_to_float(entry.attrs["timestamp"])


def iter_entries(data_file):
    """Iterate through the entries in an arf file in order of time"""
    return enumerate(sorted(data_file.values(), key=entry_time))


def find_stim_dset(entry):
    """Returns the first dataset that matches 'Network_Events.*_TEXT'"""
    rex = re.compile(r"Network_Events-.*?TEXT")
    for name in entry:
        if rex.match(name) is not None:
            log.debug("  - stim log dataset: %s", name)
            return entry[name]


def entry_metadata(entry):
    """Extracts metadata from an entry in an oeaudio-present experiment ARF file.

    Metadata are passed to open-ephys through the network events socket as a
    json-encoded dictionary. There should be one and only one metadata message
    per entry, so only the first is returned.

    """
    re_metadata = re.compile(r"metadata: (\{.*\})")
    stims = find_stim_dset(entry)
    for row in stims:
        message = row["message"].decode("utf-8")
        m = re_metadata.match(message)
        try:
            metadata = json.loads(m.group(1))
        except (AttributeError, json.JSONDecodeError):
            pass
        else:
            metadata.update(name=entry.name, sampling_rate=stims.attrs["sampling_rate"])
            return metadata


class StimulusFinder:
    """Looks up stimuli using neurobank and/or files in a local directory"""

    def __init__(self, nbank_registry_url: str, alt_base: Path | None = None):
        self.registry_url = nbank_registry_url
        self.alt_base = alt_base

    def get_durations(self, names: Iterable[str]) -> dict[str, float]:
        """Looks up durations (in s) for a sequence of stimuli. Searches
        neurobank first and then tries local directory.

        """
        output = {}
        for name, res in nbank.find_resources(*names, registry_url=self.registry_url):
            if isinstance(res, FileNotFoundError):
                if self.alt_base is None:
                    raise res
                path = (self.alt_base / name).with_suffix(".wav")
                if not path.exists():
                    raise res
            else:
                path = res
            log.debug("  - found '%s' at %s", name, path)
            with ewave.wavfile(path) as fp:
                output[name] = 1.0 * fp.nframes / fp.sampling_rate
        return output


def oeaudio_to_trials(
    data_file: h5.File,
    stim_finder: StimulusFinder,
    sync_dset: str,
    sync_thresh: float = 1.0,
    prepad: float = 1.0,
) -> Iterator[Trial]:
    """Extracts trial information from an oeaudio-present experiment ARF file

    When using oeaudio-present, a single recording is made in response to all
    the stimuli. The stimulus presentation script sends network events to
    open-ephys to mark the start and stop of each stimulus. There is typically a
    significant lag between the 'start' event and the onset of the stimulus, due
    to buffering of the audio playback. However, the oeaudio-present script will
    play a synchronization click on a second channel by default. As long as the
    user remembers to record this channel, it can be used to correct the onset
    and offset values.

    The continuous recording is broken up into trials based on the stimulus
    presentation, such that each trial encompasses one and only one stimulus.
    The `prepad` parameter specifies, in seconds, when trials begin relative to
    stimulus onset. The default is 1.0 s.

    """
    from itertools import zip_longest

    expt_start = None
    det = qs.detector(sync_thresh, 10)
    trials = []
    for entry_num, entry in iter_entries(data_file):
        log.info(" - entry: '%s'", entry.name)
        entry_start = arf.timestamp_to_float(entry.attrs["timestamp"])
        log.info(
            "  - start time: %s", arf.timestamp_to_datetime(entry.attrs["timestamp"])
        )
        if expt_start is None:
            expt_start = entry_start

        log.info("  - parsing stimulus log")
        entry_stimuli = list(oeaudio_stims(find_stim_dset(entry)))
        try:
            stim_durations = stim_finder.get_durations(
                stim.name for stim in entry_stimuli
            )
        except FileNotFoundError as err:
            raise RuntimeError(
                "unable to find a stimulus to look up duration. Was it deposited in neurobank?"
            ) from err
        log.info("    - detected %d stimuli", len(entry_stimuli))
        log.info("  - sync track: '%s'", sync_dset)
        sync = entry[sync_dset]
        sync_data = sync[:].astype("d")
        det.scale_thresh(sync_data.mean(), sync_data.std())
        stim_onsets = np.asarray(det(sync_data))
        log.info("    - detected %d clicks", stim_onsets.size)
        dset_offset = sync.attrs["offset"]
        dset_end = sync.size
        sampling_rate = sync.attrs["sampling_rate"]
        stim_sample_offset = int(dset_offset * sampling_rate)
        log.info("  - recording clock offset: %d", stim_sample_offset)

        entry_stimuli = match_clicks(entry_stimuli, stim_onsets)

        padding_samples = int(prepad * sampling_rate)
        for stim, onset, offset in zip_longest(
            entry_stimuli,
            stim_onsets,
            stim_onsets[1:],
            fillvalue=dset_end + padding_samples,
        ):
            stim_seconds = stim_durations[stim.name]
            stim_samples = int(stim_seconds * sampling_rate)
            if stim_samples > offset - onset:
                log.warning(
                    "  - WARNING: stimulus %s is longer than the duration of the trial",
                    stim,
                )
            trials.append(
                Trial(
                    entry_num,
                    onset - padding_samples,
                    offset - padding_samples,
                    stim.name,
                    onset,
                    onset + stim_samples,
                )
            )
    return trials


def match_clicks(entry_stimuli, stim_onsets):
    """Match clicks in the sync track to the stimulus onset log.

    If the number of clicks matches the number of stimuli, nothing happens. If
    there are more clicks than stimuli, this is an error. If there are more
    stimuli than clicks, attempts to match each click with the next onset,
    discarding any stimuli that don't have a matching click.

    """
    if len(entry_stimuli) == stim_onsets.size:
        logging.debug(" - Number of stimuli matches number of clicks")
        return entry_stimuli
    elif len(entry_stimuli) < stim_onsets.size:
        logging.error(
            "  - Number of stimuli (%d) is fewer than the number of clicks (%d)",
            len(entry_stimuli),
            stim_onsets.size,
        )
        raise ValueError(
            "  - Error: unable to match clicks in the sync track with the stimulus list. Either discard recording or change sync threshold."
        )
    logging.info(
        "  - Number of stimuli (%d) is greater than number of clicks (%d). Trying to repair.",
        len(entry_stimuli),
        stim_onsets.size,
    )
    # this algorithm assumes that the click comes before the message, which is
    # pretty reasonable given that network delays will be longer than analog
    # signal propagation.
    matched_stims = []
    click_times = set(stim_onsets)
    for i, stim in enumerate(entry_stimuli):
        idx = np.searchsorted(stim_onsets, stim.start)
        closest = stim_onsets[idx - 1]
        logging.info(
            "   - stim %d (start=%d) matched to click at %d", i, stim.start, closest
        )
        if closest in click_times:
            matched_stims.append(stim)
            click_times.remove(closest)
        else:
            logging.info("     - this click was already used, dropping the trial")
    return matched_stims


def assign_events_flat(events: pd.DataFrame, sampling_rate: float):
    """Assign event_times to clusters, generating a large toelis object"""
    nevents, _ = events.shape
    nclusters = events.index.unique().size
    log.info("- grouping %d spikes into %d clusters...", nevents, nclusters)
    return events.groupby("clust").apply(
        lambda df: df.time.sort_values().to_numpy() / sampling_rate * 1000.0
    )


def trials_to_pprox(trials: pd.DataFrame, sampling_rate: float):
    """Convert pandas trials to pproc"""
    for trial in trials.itertuples():
        if isinstance(trial.events, float):
            events = []
        else:
            events = (trial.events.astype("d") - trial.stimulus_start) / sampling_rate
        pproc = {
            "events": events,
            "offset": trial.stimulus_start / sampling_rate,
            "index": trial.Index,
            "interval": (
                (trial.recording_start - trial.stimulus_start) / sampling_rate,
                (trial.recording_end - trial.stimulus_start) / sampling_rate,
            ),
            "stimulus": {
                "name": trial.stimulus_name,
                "interval": (
                    0.0,
                    (trial.stimulus_end - trial.stimulus_start) / sampling_rate,
                ),
            },
            "recording": {
                "entry": trial.recording_entry,
                "start": trial.recording_start,
                "end": trial.recording_end,
            },
        }
        yield pproc


def group_spikes_script(argv=None):
    import argparse
    import os

    from dlab import __version__
    from dlab.util import json_serializable, setup_log

    version = "2025.07.28"

    p = argparse.ArgumentParser(
        description="group kilosorted spikes into pprox files based on cluster and trial"
    )
    p.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {version} (melizalab-tools {__version__})",
    )
    p.add_argument("--debug", help="show verbose log messages", action="store_true")
    nbank.add_registry_argument(p)
    p.add_argument(
        "--dry-run",
        help="do everything except write the output files",
        action="store_true",
    )
    p.add_argument(
        "--sync",
        default="sync",
        help="name of channel with synchronization signal (default '%(default)s')",
    )
    p.add_argument(
        "--sync-thresh",
        default=30.0,
        type=float,
        help="threshold (z-score) for detecting sync clicks (default %(default)0.1f)",
    )
    p.add_argument(
        "--prepad",
        type=float,
        default=1.0,
        help="sets trial start time relative to stimulus onset (default %(default)0.1f s)",
    )
    p.add_argument(
        "--toelis",
        action="store_true",
        help="output toelis instead of pprox. one file will be generated for "
        "the entire recording (including multiunits)",
    )
    p.add_argument(
        "--cluster",
        "-c",
        help="only save data for the specified clusters (as comma-separated list)",
        type=lambda s: [int(item) for item in s.split(",")],
    )
    p.add_argument(
        "--output",
        "-o",
        type=Path,
        default=".",
        help="directory to output pprox files (default current directory)",
    )
    p.add_argument(
        "--mua",
        action="store_true",
        help="save multiunit clusters along with single units",
    )
    p.add_argument(
        "--artifact-reject-thresh",
        type=float,
        default=6.0,
        help="threshold for rejecting artifact spikes (default %(default).1f; max absolute amplitude"
        " more than x times max absolute amplitude of the mean spike)",
    )
    p.add_argument(
        "--no-waveforms",
        "-W",
        action="store_true",
        help="if set, do not save representative waveforms from each unit's main channel in an hdf5 file",
    )
    p.add_argument(
        "--waveform-pre-peak",
        type=float,
        default=2.0,
        help="samples before the spike to keep (default %(default).1f ms)",
    )
    p.add_argument(
        "--waveform-post-peak",
        type=float,
        default=5.0,
        help="samples after the spike to keep (default %(default).1f ms)",
    )
    p.add_argument(
        "--local-stim-dir",
        type=Path,
        help="DEBUG/TESTING ONLY. Search this directory for stimulus files.",
    )
    p.add_argument("recording", type=Path, help="path of ARF recording file")
    p.add_argument(
        "sortdir",
        type=Path,
        help="kilosort output directory. Needs to contain 'spike_times.npy', 'spike_clusters.npy',"
        " 'cluster_info.tsv', and 'temp_wh.dat'",
    )
    args = p.parse_args(argv)
    setup_log(args.debug)
    os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"
    log.info("- %s version %s", p.prog, version)

    log.info("- using neurobank registry at %s", args.registry_url)
    log.info("- using stimulus times from %s", args.recording)
    try:
        resource_info = nbank.describe(args.registry_url, args.recording.stem)
        recording_name = resource_info["name"]
        resource_url = nbank.registry.full_url(args.registry_url, recording_name)
        log.info("  - registered at %s", resource_url)
    except TypeError:
        if args.debug:
            resource_info = {"metadata": {}}
            recording_name = args.recording.stem
            resource_url = "(debug)"
            log.warning(
                "  - warning: recording has not been deposited, proceeding anyway in debug mode"
            )
        else:
            log.error("  - error: recording must be deposited in neurobank")
            p.exit(-1)

    if args.local_stim_dir is not None:
        log.info("  - using %s as fallback for looking up stimuli", args.local_stim_dir)
    stim_finder = StimulusFinder(args.registry_url, args.local_stim_dir)

    log.info("- kilosort output directory: %s", args.sortdir)
    timefile = args.sortdir / "spike_times.npy"
    clustfile = args.sortdir / "spike_clusters.npy"
    infofile = args.sortdir / "cluster_info.tsv"
    log.info("  - spike times: %s", timefile)
    log.info("  - spike clusters: %s", clustfile)
    events = pd.DataFrame(
        {"time": np.load(timefile).squeeze(), "clust": np.load(clustfile)},
    )
    log.info("  - cluster info: %s", infofile)
    info = pd.read_csv(infofile, sep="\t", index_col=0)
    recfile = args.sortdir / "temp_wh.dat"
    params = read_kilo_params(args.sortdir / "params.py")
    recording = np.memmap(recfile, mode="c", dtype=params["dtype"])
    recording = np.reshape(
        recording, (recording.size // params["nchannels"], params["nchannels"])
    )
    nsamples, nchannels = recording.shape
    log.info("  - filtered recording: %s", recfile)
    log.info("    - %d samples, %d channels", nsamples, nchannels)
    if args.cluster is not None:
        log.info("- only analyzing clusters: %s", args.cluster)
        events = events[events.cluster == args.cluster]

    # find duplicates - this is rare but needs to be caught
    duplicates = events.duplicated()
    if duplicates.any():
        dupl_clusts = events[duplicates].clust
        log.warning(
            "  - warning: removing spikes with duplicate times from clusters %s",
            ",".join(str(c) for c in dupl_clusts),
        )
        events = events[~duplicates]

    events.set_index("clust", inplace=True)
    if args.toelis:
        clusters = assign_events_flat(events, params["sampling_rate"])
        outfile = (args.output / recording_name).with_suffix(".toe_lis")
        if not args.dry_run:
            with open(outfile, "w") as ofp:
                toelis.write(ofp, clusters)
            log.info("- saved %d spikes to '%s'", toelis.count(clusters), outfile)
        return

    if args.recording.is_file():
        datafile = args.recording
    else:
        datafile = nbank.find_resource(
            str(args.recording), registry_url=nbank.default_registry
        )
    log.info("- splitting '%s' into trials:", datafile)
    with h5.File(datafile, "r") as afp:
        trials = pd.DataFrame(
            oeaudio_to_trials(
                afp, stim_finder, args.sync, args.sync_thresh, args.prepad
            )
        )
        entry_attrs = tuple(entry_metadata(e) for _, e in iter_entries(afp))

    # this pandas magic sorts the events by cluster and trial
    log.info("- sorting events into trials:")
    events["trial"] = trials.recording_start.searchsorted(events.time, side="left") - 1

    total_spikes = 0
    total_clusters = 0
    good_clust_types = ("good",)
    if args.mua:
        good_clust_types += ("mua",)
    for clust_id, cluster in events.groupby("clust"):
        clust_info = info.loc[clust_id]
        clust_type = clust_info["group"]
        n_spikes = len(cluster)
        if clust_type not in good_clust_types:
            log.info(
                "  - cluster %d (%d spikes, %s) -> skipped",
                clust_id,
                n_spikes,
                clust_type,
            )
            continue
        log.info(
            "  ✓ cluster %d (%d spikes, %s)",
            clust_id,
            n_spikes,
            clust_type,
        )
        # remove artifact spikes
        n_before = int(args.waveform_pre_peak * params["sampling_rate"] / 1000)
        n_after = int(args.waveform_post_peak * params["sampling_rate"] / 1000)
        spikes = cluster[
            (cluster.time > n_before) & (cluster.time < (nsamples - n_after))
        ]
        waveforms = qs.peaks(
            recording[:, clust_info["ch"]],
            spikes.time,
            n_before=n_before,
            n_after=n_after,
        )
        mean_spike = waveforms.mean(0)
        included = np.abs(waveforms).max(-1) < (
            np.abs(mean_spike).max(-1) * args.artifact_reject_thresh
        )
        n_included = included.sum()
        if n_included == 0:
            log.warning("   - all spikes marked as artifacts (sorting error?)")
            continue
        elif n_included < n_spikes:
            cluster = cluster[included]
            waveforms = waveforms[included]
            log.info("    - %d artifact spike(s) excluded", n_spikes - n_included)
        # aggregate spikes by trial and left join to trial information table
        # - empty trials will be nan
        clust_trials = trials.join(
            cluster.groupby("trial")
            .apply(lambda x: x.time.to_numpy(), include_groups=False)
            .rename("events")
        )
        total_spikes += n_spikes
        total_clusters += 1
        outfile = args.output / f"{recording_name}_c{clust_id}.pprox"
        log.info(
            "    - %d spikes -> %s",
            n_included,
            outfile,
        )
        clust_trials = pprox.from_trials(
            trials_to_pprox(clust_trials, params["sampling_rate"]),
            schema=pprox._stimtrial_schema,
            recording=resource_url,
            processed_by=[f"{p.prog} {version}"],
            kilosort_amplitude=clust_info["Amplitude"],
            kilosort_contam_pct=clust_info["ContamPct"],
            kilosort_source_channel=clust_info["ch"],
            kilosort_probe_depth=clust_info["depth"],
            kilosort_n_spikes=clust_info["n_spikes"],
            entry_metadata=entry_attrs,
            **resource_info["metadata"],
        )
        if not args.dry_run:
            with open(outfile, "w") as ofp:
                json.dump(clust_trials, ofp, default=json_serializable)
            if not args.no_waveforms:
                outfile = args.output / (outfile.stem + "_spikes.h5")
                log.info(
                    "    - waveforms on channel %d -> %s",
                    clust_info["ch"],
                    outfile,
                )
                save_waveforms(
                    outfile,
                    SpikeWaveforms(
                        waveforms,
                        cluster.time.to_numpy(),
                        params["sampling_rate"],
                        n_before,
                    ),
                    recording=resource_url,
                    processed_by=f"{p.prog} {version}",
                    kilosort_amplitude=clust_info["Amplitude"],
                    kilosort_contam_pct=clust_info["ContamPct"],
                    kilosort_source_channel=clust_info["ch"],
                    kilosort_probe_depth=clust_info["depth"],
                    kilosort_n_spikes=clust_info["n_spikes"],
                )

    log.info(
        "- a total of %d spikes were assigned to %d clusters",
        total_spikes,
        total_clusters,
    )


if __name__ == "__main__":
    group_spikes_script()



# -*- mode: python -*-
"""Utility functions for scripts and modules"""

import argparse
import logging
from functools import singledispatch

import numpy as np


def setup_log(debug=False):
    logging.basicConfig(
        format="%(message)s", level=logging.DEBUG if debug else logging.INFO
    )
    # suppress info messages from httpx
    logging.getLogger("httpx").setLevel(logging.DEBUG if debug else logging.WARNING)


class ParseKeyVal(argparse.Action):
    """argparse action for parsing -k key=value arguments

    Example: p.add_argument("-k", action=ParseKeyVal, default=dict(),
                            metavar="KEY=VALUE", dest="metadata")
    """

    def parse_value(self, value):
        import ast

        try:
            return ast.literal_eval(value)
        except (ValueError, SyntaxError):
            return value

    def __call__(self, parser, namespace, arg, option_string=None):
        kv = getattr(namespace, self.dest)
        if kv is None:
            kv = dict()
        if not arg.count("=") == 1:
            raise ValueError(f"{arg} argument badly formed; needs key=value")
        else:
            key, val = arg.split("=")
            kv[key] = self.parse_value(val)
        setattr(namespace, self.dest, kv)


@singledispatch
def json_serializable(val):
    """Serialize a value for the json module."""
    return str(val)


@json_serializable.register(np.generic)
def __js_numpy(val):
    """Used if *val* is an instance of a numpy scalar."""
    return val.item()


@json_serializable.register(np.ndarray)
def __js_numpy_arr(arr):
    """Used if *arr* is an instance of a numpy array."""
    return arr.tolist()


def all_same(seq):
    """If all elements of seq are equal, returns the value, otherwise None"""
    it = iter(seq)
    first = next(it)
    for e in it:
        if e != first:
            return None
    return first



# -*- mode: python -*-
"""Functions for interfacing with the neurobank repository"""

import concurrent.futures
import json
import logging
from collections.abc import Iterator
from pathlib import Path
from urllib.parse import urlparse

from httpx import Client, HTTPStatusError, NetRCAuth
from nbank import registry, util
from nbank.archive import resolve_extension
from nbank.core import deposit, describe, describe_many
from nbank.registry import log_error

from dlab import cache

log = logging.getLogger(__name__)
default_registry = registry.default_registry()
default_auth = NetRCAuth(None)

MaybeResourcePath = tuple[str, Path | FileNotFoundError]
MaybeResourceMetadata = tuple[str, dict | FileNotFoundError]


def find_resources(
    *resource_ids: str,
    registry_url: str | None = default_registry,
    alt_base: Path | str | None = None,
    no_download: bool = False,
) -> Iterator[MaybeResourcePath]:
    """Locate resources using neurobank or a local directory.

    This function will try to locate resources from the following locations:
    - a local directory (alt_base, if set),
    - a local neurobank archive (using registry_url and alt_base, if provided),
    - a local cache of previously fetched resources
    - a remote HTTP archive (using registry_url, caching the file for local access later)

    Yields results as they become available. The result for each requested id is
    a Path if the resource was successfully located, or exist, or a
    FileNotFoundError if the resource cannot be located.

    """
    to_locate = set(resource_ids)
    if alt_base is not None:
        path = Path(alt_base)
        for resource_id in resource_ids:
            stem = path / resource_id
            try:
                yield (resource_id, resolve_extension(stem))
                to_locate.remove(resource_id)
                log.debug("%s: found in alt_base", resource_id)
            except FileNotFoundError:
                pass
    if len(to_locate) == 0:
        return
    url, query = registry.get_locations_bulk(registry_url, to_locate)
    with Client() as client, concurrent.futures.ThreadPoolExecutor() as executor:
        response = util.query_registry_bulk(client, url, query)
        future_to_name = {
            executor.submit(
                fetch_resource,
                client,
                resource,
                alt_base=alt_base,
                no_download=no_download,
            ): resource["name"]
            for resource in response
        }
        for future in concurrent.futures.as_completed(future_to_name):
            resource_id = future_to_name[future]
            try:
                result = future.result()
            except FileNotFoundError as err:
                result = err
            to_locate.remove(resource_id)
            yield (resource_id, result)
    for resource_id in to_locate:
        yield (resource_id, FileNotFoundError(f"no such resource {resource_id}"))


def find_resource(
    resource_id: str,
    *,
    registry_url: str = default_registry,
    alt_base: Path | str | None = None,
    no_download: bool = False,
) -> Path:
    """Locate a resource using neurobank. This is a convenience wrapper for find_resources"""
    for _, result in find_resources(
        resource_id,
        registry_url=registry_url,
        alt_base=alt_base,
        no_download=no_download,
    ):
        if isinstance(result, Path):
            return result
        else:
            raise result


def fetch_resource(
    client: Client,
    resource: dict,
    *,
    alt_base: Path | str | None = None,
    no_download: bool = False,
) -> Path:
    """Fetch a resource.

    Given a sequence of locations where a resource could be found, this function
    will try to locate the file, prioritizing local archives.

    """
    resource_id = resource["name"]
    filename = resource.get("filename", resource_id)
    for location in resource["locations"]:
        loc = util.parse_location(location, alt_base=alt_base, http_session=client)
        if hasattr(loc, "path"):
            log.debug("- found in local repository: %s", location)
            return loc.path
        elif hasattr(loc, "url"):
            target = cache.locate(
                filename, Path(urlparse(loc.url).netloc) / "resources"
            )
            if target.exists():
                log.debug("- found in local cache: %s", target)
                return target
            if no_download:
                log.debug("- found at %s but configured not to download", loc.url)
                continue
            log.debug("- fetching from %s", loc.url)
            try:
                return loc.fetch(target)
            except HTTPStatusError as err:
                log.warning(
                    "%s: failed to retrieve from %s (http status %d)",
                    resource_id,
                    loc.url,
                    err.response.status_code,
                )
    raise FileNotFoundError(
        "resource not found in local archive, cache, or downloadable remote"
    )


def describe_resources(
    *resource_ids: str,
    registry_url: str | None = default_registry,
    alt_base: Path | str | None = None,
) -> Iterator[MaybeResourceMetadata]:
    """Fetch resource metadata using neurobank or a local directory.

    This function will try to load resource metadta from the following locations:
    - a local directory (alt_base, if set),
    - a remote neurobank registry

    Yields results as they become available. The result for each requested id is
    a dict if the metadata was successfully loaded or FileNotFoundError if the resource cannot be located.

    """
    to_locate = set(resource_ids)
    if alt_base is not None:
        alt_base = Path(alt_base)
        for resource_id in resource_ids:
            path = (alt_base / resource_id).with_suffix(".json")
            if path.exists():
                yield (resource_id, json.loads(path.read_text()))
                to_locate.remove(resource_id)
                log.debug("%s: found in alt_base", resource_id)
    if len(to_locate) == 0:
        return
    try:
        for result in describe_many(registry_url, *to_locate):
            yield (result["name"], result)
    except Exception:
        for name in to_locate:
            yield (
                name,
                FileNotFoundError("not found in local dir, unable to access registry"),
            )


def add_registry_argument(parser, dest="registry_url"):
    """Add a registry argument to an argparse parser"""
    parser.add_argument(
        "-r",
        "--registry",
        dest=dest,
        help="URL of the registry service. "
        f"Default is to use the environment variable '{registry._env_registry}'",
        default=default_registry,
    )


def main(argv=None):
    import argparse

    from dlab.util import setup_log

    p = argparse.ArgumentParser(description="locate neurobank resources ")
    p.add_argument("--debug", help="show verbose log messages", action="store_true")
    add_registry_argument(p)
    p.add_argument(
        "-b",
        "--base",
        type=Path,
        help="set an alternative base directory to search for local resources",
    )
    p.add_argument(
        "--clear-cache",
        action="store_true",
        help="clear the contents of the local cache",
    )
    p.add_argument("id", help="identifier(s) of the resource(s) to locate", nargs="*")
    args = p.parse_args(argv)
    setup_log(args.debug)

    if args.clear_cache:
        cache.clear(urlparse(args.registry_url).netloc)

    if len(args.id) > 0:
        for name, location in find_resources(
            *args.id, registry_url=args.registry_url, alt_base=args.base
        ):
            logging.info("%s: %s", name, location)


__all__ = [
    "HTTPStatusError",
    "add_registry_argument",
    "default_auth",
    "default_registry",
    "deposit",
    "describe",
    "describe_resources",
    "fetch_resource",
    "find_resource",
    "find_resources",
    "log_error",
]

if __name__ == "__main__":
    main()



# -*- mode: python -*-
"""Signal processing functions"""

import dataclasses
from pathlib import Path

import ewave
import numpy as np


@dataclasses.dataclass
class Signal:
    samples: np.ndarray
    sampling_rate: float  # in Hz
    name: str | None = None
    duration: float = dataclasses.field(init=False)  # in s
    dBFS: float = dataclasses.field(init=False)

    def __post_init__(self):
        self.duration = 1.0 * self.samples.size / self.sampling_rate
        self.dBFS = dBFS(self.samples)

    def write_wav(self, fname: str | Path, dtype: str = "h"):
        with ewave.open(fname, mode="w", dtype=dtype, sampling_rate=self.sampling_rate) as fp:
            fp.write(self.samples)


def dBFS(samples: np.ndarray) -> float:
    """Returns the RMS level of samples, in dB FS (relative to 1.0)"""
    rms = np.sqrt(np.mean(samples * samples))
    return 20 * np.log10(rms) + 3.0103


def peak(samples: np.ndarray) -> float:
    """Returns the peak level of samples, in dB FS (relative to 1.0)"""
    absmax = np.amax(np.absolute(samples))
    return 20 * np.log10(absmax)


def resample(signal: Signal, target: float) -> Signal:
    """Resample the signal to target rate (in Hz)"""
    import samplerate
    if signal.sampling_rate == target:
        return signal
    ratio = 1.0 * target / signal.sampling_rate
    # NB: this silently converts data to float32
    data = samplerate.resample(signal.samples, ratio, "sinc_best")
    return Signal(
        name=signal.name,
        samples=data,
        sampling_rate=target,
    )


def hp_filter(signal: Signal, cutoff_Hz: float, order: int = 12) -> Signal:
    """Highpass filter the signal to remove DC and low-frequency noise"""
    from scipy.signal import butter, sosfilt

    sos = butter(
        order, cutoff_Hz, fs=signal.sampling_rate, btype="highpass", output="sos"
    )
    filtered = sosfilt(sos, signal.samples)
    return Signal(
        name=signal.name,
        samples=filtered,
        sampling_rate=signal.sampling_rate,
    )


def rescale(signal: Signal, target: float) -> Signal:
    """Rescale the signal to a target dBFS"""
    scale = 10 ** ((target - signal.dBFS) / 20)
    return Signal(
        name=signal.name,
        samples=signal.samples * scale,
        sampling_rate=signal.sampling_rate,
    )


def ramp_signal(signal: Signal, duration_s: float = 0.002) -> Signal:
    """Apply a squared cosine ramp to the beginning and end of signal."""
    from numpy import cos, linspace, pi, sin

    s = signal.samples.copy()
    n = int(duration_s * signal.sampling_rate)
    t = linspace(0, pi / 2, n)
    s[:n] *= sin(t) ** 2
    s[-n:] *= cos(t) ** 2
    return Signal(
        name=signal.name,
        samples=s,
        sampling_rate=signal.sampling_rate,
    )


def smoothing_kernel(
    name: str, bandwidth: float, dt: float
) -> tuple[np.ndarray, np.ndarray]:
    """
    Generate a (normalized) smoothing kernel function with specified bandwidth.

    name:      the name of the kernel. can be anything supported
               by scipy.signal.get_window, plus the following functions:
        epanech, biweight, triweight, cosinus, exponential

    bandwidth: the bandwidth of the kernel, in units of dt
    dt:        the time resolution of the kernel

    Returns:
    window:    the window function, normalized such that sum(w)*dt = 1.0
    grid:      the time support of the window, centered around 0.0

    From matlab code by Zhiyi Chi
    """
    if name in ("normal", "gaussian"):
        D = 3.75 * bandwidth
    elif name == "exponential":
        D = 4.75 * bandwidth
    else:
        D = bandwidth

    N = int(np.floor(D / dt))  # number of grid points in half the support
    G = (np.arange(1, 2 * N + 2) - 1 - N) * dt  # grid support
    xv = G / bandwidth

    if name in ("gaussian", "normal"):
        W = np.exp(-xv * xv / 2)
    elif name == "exponential":
        xv = np.minimum(xv, 0)
        W = np.absolute(xv) * np.exp(xv)
    elif name in ("biweight", "quartic"):
        W = np.maximum(0, 1 - xv * xv) ** 2
    elif name == "triweight":
        W = np.maximum(0, 1 - xv * xv) ** 3
    elif name == "cosinus":
        W = np.cos(np.pi * xv / 2) * (np.absolute(xv) <= 1)
    elif name == "epanech":
        W = np.maximum(0, 1 - xv * xv)
    elif name == "hanning":
        W = np.hanning(1 + 2 * N)
    else:
        from scipy.signal import get_window

        W = get_window(name, 1 + 2 * N)

    return W / (W.sum() * dt), G


# lifted from https://github.com/endolith/waveform_analysis


def ABC_weighting(curve="A"):
    """
    Design an analog weighting filter with A, B, or C curve.

    Returns zeros, poles, gain of the filter.
    """
    from scipy.signal import freqs, zpk2tf

    if curve not in "ABC":
        raise ValueError(f"Curve type {curve} not supported")

    # ANSI S1.4-1983 C weighting
    #    2 poles on the real axis at "20.6 Hz" HPF
    #    2 poles on the real axis at "12.2 kHz" LPF
    #    -3 dB down points at "10^1.5 (or 31.62) Hz"
    #                         "10^3.9 (or 7943) Hz"
    #
    # IEC 61672 specifies "10^1.5 Hz" and "10^3.9 Hz" points and formulas for
    # derivation.  See _derive_coefficients()

    z = [0, 0]
    p = [
        -2 * np.pi * 20.598997057568145,
        -2 * np.pi * 20.598997057568145,
        -2 * np.pi * 12194.21714799801,
        -2 * np.pi * 12194.21714799801,
    ]
    k = 1

    if curve == "A":
        # ANSI S1.4-1983 A weighting =
        #    Same as C weighting +
        #    2 poles on real axis at "107.7 and 737.9 Hz"
        #
        # IEC 61672 specifies cutoff of "10^2.45 Hz" and formulas for
        # derivation.  See _derive_coefficients()

        p.append(-2 * np.pi * 107.65264864304628)
        p.append(-2 * np.pi * 737.8622307362899)
        z.append(0)
        z.append(0)

    elif curve == "B":
        # ANSI S1.4-1983 B weighting
        #    Same as C weighting +
        #    1 pole on real axis at "10^2.2 (or 158.5) Hz"

        p.append(-2 * np.pi * 10**2.2)  # exact
        z.append(0)

    # TODO: Calculate actual constants for this
    # Normalize to 0 dB at 1 kHz for all curves
    b, a = zpk2tf(z, p, k)
    k /= abs(freqs(b, a, [2 * np.pi * 1000])[1][0])

    return np.asarray(z), np.asarray(p), k


def A_weighting(fs: float, output="ba"):
    """
    Design of a digital A-weighting filter.

    Designs a digital A-weighting filter for
    sampling frequency `fs`.
    Warning: fs should normally be higher than 20 kHz. For example,
    fs = 48000 yields a class 1-compliant filter.

    Parameters
    ----------
    fs : float
        Sampling frequency
    output : {'ba', 'zpk', 'sos'}, optional
        Type of output:  numerator/denominator ('ba'), pole-zero ('zpk'), or
        second-order sections ('sos'). Default is 'ba'.

    """
    from scipy.signal import bilinear_zpk, zpk2sos, zpk2tf

    z, p, k = ABC_weighting("A")

    # Use the bilinear transformation to get the digital filter.
    z_d, p_d, k_d = bilinear_zpk(z, p, k, fs)

    if output == "zpk":
        return z_d, p_d, k_d
    elif output in {"ba", "tf"}:
        return zpk2tf(z_d, p_d, k_d)
    elif output == "sos":
        return zpk2sos(z_d, p_d, k_d)
    else:
        raise ValueError(f"'{output}' is not a valid output form.")



# -*- mode: python -*-
import numpy as np
import pytest

from dlab import signal

binsize = 0.01
bandwidth = 0.1
kernels = [
    "gaussian",
    "exponential",
    "biweight",
    "triweight",
    "cosinus",
    "epanech",
    "hanning",
]


@pytest.fixture(params=kernels)
def conv_kernel(request):
    return signal.smoothing_kernel(request.param, bandwidth, binsize)


@pytest.fixture
def test_signal():
    np.random.seed(1028)
    samples = np.random.randn(48000)
    samples *= 0.1 / samples.std()
    return signal.Signal(samples, 48000)


def test_kernel_scale(conv_kernel):
    k, kt = conv_kernel
    assert k.size == kt.size
    assert k.sum() * binsize == pytest.approx(1.0)


def test_signal_attributes(test_signal):
    expected_dBFS = -20 + 3.0103
    assert test_signal.duration == 1.0
    assert np.abs(test_signal.dBFS - expected_dBFS) < 0.001


def test_signal_resample(test_signal):
    resampled = signal.resample(test_signal, target=24000)
    assert resampled.sampling_rate == 24000
    assert resampled.duration == test_signal.duration
    # resampling will change the scale
    # assert resampled.dBFS == test_signal.dBFS


def test_signal_rescale(test_signal):
    target_dBFS = -40
    rescaled = signal.rescale(test_signal, target=target_dBFS)
    assert rescaled.sampling_rate == test_signal.sampling_rate
    assert rescaled.duration == test_signal.duration
    assert np.abs(rescaled.dBFS - target_dBFS) < 0.001


def test_hp_filter(test_signal):
    # add a DC offset
    test_signal.samples += 1
    assert test_signal.samples.mean() > 0
    filtered = signal.hp_filter(test_signal, cutoff_Hz=100)
    assert filtered.sampling_rate == test_signal.sampling_rate
    assert filtered.duration == test_signal.duration
    assert np.abs(filtered.samples.mean()) < 0.001



# -*- mode: python -*-
"""Local cache, primarily for storing neurobank resources"""

import logging
from pathlib import Path
from shutil import rmtree

import appdirs

APP_NAME = "dlab"
APP_AUTHOR = "melizalab"
user_dir = appdirs.user_cache_dir(APP_NAME, APP_AUTHOR)
log = logging.getLogger(__name__)


def locate(name: Path | str, subdir: Path | str) -> Path:
    """Return Path for a cached resource, creating subdir in the cache if needed"""
    cache_dir = Path(user_dir) / subdir
    cache_dir.mkdir(parents=True, exist_ok=True)
    return cache_dir / name


def clear(subdir: str) -> None:
    """Clear the contents of the cache"""
    cache_dir = Path(user_dir) / subdir
    log.info("clearing local cache dir %s", cache_dir)
    rmtree(cache_dir, ignore_errors=True)



# -*- mode: python -*-
"""Extract songs from the neurobank repository to use as stimuli.

This script uses a YAML file to control its behavior. The file needs to have the following format:

- name: O103
  source: O103_song
  dataset: entry_00004/pcm_000
  interval_ms: [90.0, 2185.0]
- name: O84
  source: O84_song
  dataset: entry_00004/pcm_000
  interval_ms: [2970.0, 4922.0]

For each item (denoted by a block starting with `-`), the script will open the
ARF file corresponding to `source` and extract the samples in the interval
`interval_ms` from the dataset denoted by `dataset`. The segments are resampled
to the target sampling rate, passed through a highpass filter, rescaled to the
target amplitude, and then saved in WAVE format. The parameters of the
resampling, filtering, and rescaling are set using commandline options. After
you're happy with the output, use the `--deposit` flag can be used to deposit
the WAVE files in neurobank along with metadata.

"""

import logging
import os
from collections.abc import Sequence
from pathlib import Path

import ewave
import numpy as np

from dlab import neurobank as nbank
from dlab.signal import Signal, hp_filter, resample, rescale

# disable locking - neurobank archive is probably on an NFS share
os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"
log = logging.getLogger("dlab")


def get_interval(path: Path, dataset: str, interval_ms: Sequence[float]) -> Signal:
    import h5py as h5

    with h5.File(path, "r") as fp:
        dset = fp[dataset]
        sampling_rate = dset.attrs["sampling_rate"]
        start, stop, *_rest = (int(t * sampling_rate / 1000) for t in interval_ms)
        data = dset[slice(start, stop)].astype("float32")
        return Signal(samples=data, sampling_rate=sampling_rate)


def script(argv=None):
    import argparse

    import yaml

    from dlab import __version__
    from dlab.util import setup_log

    script_version = "2025.07.28"

    p = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    p.add_argument("--debug", help="show verbose log messages", action="store_true")
    p.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {script_version} (melizalab-tools {__version__})",
    )
    nbank.add_registry_argument(p)
    p.add_argument(
        "--rate",
        type=int,
        default=44100,
        help="sampling rate for the output files (default: %(default)d Hz)",
    )
    p.add_argument(
        "--dBFS",
        type=float,
        default=-20,
        help="target amplitude for the output files (default: %(default)s dBFS)",
    )
    p.add_argument(
        "--highpass",
        type=float,
        default=300,
        help="cutoff frequency for a highpass butterworth filter to apply between resampling and rescaling (default: %(default)s Hz)",
    )
    p.add_argument(
        "--filter-order",
        type=int,
        default=10,
        help="order for the butterworth highpass filter (default: %(default)d)",
    )
    p.add_argument(
        "--dtype",
        default="int16",
        help="specify data type of the output sound file (default: %(default)s)",
    )
    p.add_argument(
        "--deposit",
        type=Path,
        help="deposit files in neurobank archive (requires write access to registry)",
    )
    p.add_argument("songs", type=Path, help="YAML file with songs to extract")
    args = p.parse_args(argv)
    setup_log(args.debug)

    with open(args.songs) as fp:
        songs = yaml.safe_load(fp)

    for song in songs:
        log.info(f"{song['source']}:")
        try:
            path = nbank.find_resource(song["source"], registry_url=args.registry_url)
        except FileNotFoundError:
            path = Path(song["source"])
            if path.exists():
                log.debug(" - using local path")
            else:
                log.warning(" - unable to find resource, skipping")
                continue
        log.info(" - using %s", path)

        song_data = get_interval(path, song["dataset"], song["interval_ms"])
        song_data.name = song["name"]
        log.info(
            f" - read from dataset {song['dataset']}: {song_data.samples.size} samples, RMS {song_data.dBFS:.2f} dBFS"
        )

        song_data = resample(song_data, args.rate)
        log.info(f" - adjusted sampling rate to {song_data.sampling_rate}")
        if args.highpass:
            song_data = hp_filter(song_data, args.highpass, args.filter_order)
            log.info(
                f" - highpass with cutoff of {args.highpass} Hz. RMS is now {song_data.dBFS:.2f} dBFS"
            )

        song_data = rescale(song_data, args.dBFS)
        absmax = np.amax(np.absolute(song_data.samples))
        log.info(f" - adjusted RMS to {song_data.dBFS:.2f} dBFS (peak is {absmax:.3f})")

        out_file = Path(song_data.name + ".wav")
        with ewave.open(
            out_file,
            mode="w",
            sampling_rate=song_data.sampling_rate,
            dtype=args.dtype,
        ) as fp:
            fp.write(song_data.samples)
        log.info(f" - wrote data to {out_file}")

        if args.deposit:
            metadata = {
                "source_resource": song["source"],
                "source_dataset": song["dataset"],
                "source_interval_ms": song["interval_ms"],
                "dBFS": song_data.dBFS,
                "created_by": f"{p.prog} {script_version}",
            }
            if args.highpass:
                metadata.update(
                    highpass_cutoff=args.highpass,
                    highpass_order=args.filter_order,
                    highpass_filter="butterworth",
                )
            try:
                _res = next(
                    nbank.deposit(
                        args.deposit,
                        (out_file,),
                        dtype="vocalization-wav",
                        auth=nbank.default_auth,
                        hash=True,
                        auto_id=True,
                        **metadata,
                    )
                )
            except nbank.HTTPStatusError as err:
                log.warning(f" ✗ unable to deposit {out_file} to {args.deposit}")
                nbank.log_error(err)
        log.info("")


if __name__ == "__main__":
    script()
